package ca.uwaterloo.cs.jgrok.lib;

import java.util.*;

import ca.uwaterloo.cs.jgrok.interp.Type;

public class FunctionLib {
    private String name = null;
    private HashMap<String,Object> funEntries;
    
    public FunctionLib(String name) {
        this.name = name;
        this.funEntries = new HashMap<String,Object>();
        
        try {
            register("examples.ExampleFunction");
            register("tt2", "examples.ExampleFunction");
        } catch(RegistrationException e) {
            System.err.println("FunctionLib: Unable to register " + e.getMessage());
        }
    }
    
    public String getName() {
        return this.name;
    }
    
    public void setName(String name) {
        if(name == null) {
            throw new NullPointerException(name);
        }
        
        if(this == FunctionLibManager.JGrokLib) {
            if(!FunctionLibManager.JGrokLibName.equals(name)) {
                throw new IllegalArgumentException(name);
            }
            return;
        }
        
        this.name = name;
    }
    
    public static FunctionLib rootLib() {
        return FunctionLibManager.JGrokLib;
    }
    
    public boolean register(Function f) {
        if(f == null) return false;
        
        Entry aEntry = new Entry(f);
        putEntry(f.getName(), aEntry);
        return true;
    }
    
    public boolean register(String className)
        throws RegistrationException {
        if(className == null) return false;
        
        try {
            Class<? extends Function> defClass;
            defClass = Class.forName(className).asSubclass(Function.class);
            Function func = defClass.newInstance();
            return register(func);
        } catch(Exception e) {
            throw new RegistrationException(e.getMessage());
        }
    }
    
    public boolean register(String alias, String className)
        throws RegistrationException {
        if(alias == null || className == null) return false;
        
        try {
            Class<? extends Function> defClass;
            defClass = Class.forName(className).asSubclass(Function.class);
            Function func = defClass.newInstance();
            func.setName(alias);
            return register(func);
        } catch(Exception e) {
            throw new RegistrationException(e.getMessage());
        }
    }
    
    public String[] getFunctionNames() {
        Set<String> keySet = funEntries.keySet();
        String[] names = new String[keySet.size()];
        keySet.toArray(names);
        return names;
    }
    
    private void putEntry(String key, Entry entry) {
        Object o = funEntries.get(key);
        if(o == null) {
            funEntries.put(key, entry);
        } else if(o instanceof Entry) {
            EntryContainer econ;
            econ = new EntryContainer();
            funEntries.put(key, econ);
            econ.add((Entry)o);
            econ.add(entry);
        } else {
            ((EntryContainer)o).add(entry);
        }
    }
    
    public Function[] find(String funcName) {
        Object o;
        Entry entry;
        Function[] result;
        
        o = funEntries.get(funcName);
        
        if(o == null) {
            result = new Function[0];
        } else if(o instanceof Entry) {
            entry = (Entry)o;
            result = new Function[1];
            result[0] = entry.exeFunction;
        } else {
            ArrayList<Entry> list;
            list = ((EntryContainer)o).list;
            result = new Function[list.size()];
            
            for(int i = 0; i < list.size(); i++) {
                entry = list.get(i);
                result[i] = entry.exeFunction;
            }
        }
        
        return result;
    }
    
    public Function find(String funcName, Class<?>[] argTypes) {
        Object o;
        Entry entry, specific;
        
        o = funEntries.get(funcName);
        
        if(o == null) {
            return null;
        } else if(o instanceof Entry) {
            entry = (Entry)o;
            if(entry.anyParams()) {
                return entry.exeFunction;
            } else if(entry.endsWithAnyParams()) {
                Class<?>[] extendedParamTypes = null;
                extendedParamTypes = entry.getExtendedParamTypes(argTypes);
                if (extendedParamTypes != null &&
                    Type.distanceTo(argTypes, extendedParamTypes) >= 0) {
                    return entry.exeFunction;
                }
            } else {
                if(Type.distanceTo(argTypes, entry.paramTypes) >= 0) {
                    return entry.exeFunction;
                } 
            }
        } else {
            int distNew;
            int distOld = 1000;
            Entry anyEntry = null;
            Entry anyBetter = null;
            ArrayList<Entry> list = ((EntryContainer)o).list;
            
            specific = null;
            for(int i = 0; i < list.size(); i++) {
                entry = list.get(i);
                
                if(entry.anyParams()) {
                    anyEntry = entry;
                    continue;
                }
                
                if(entry.endsWithAnyParams()) {
                    Class<?>[] extendedParamTypes = null;
                    extendedParamTypes = entry.getExtendedParamTypes(argTypes);
                    if (extendedParamTypes != null &&
                        Type.distanceTo(argTypes, extendedParamTypes) >= 0) {
                        anyBetter = entry;
                        continue;
                    }
                }
                
                distNew = Type.distanceTo(argTypes, entry.paramTypes);
                if(distNew >= 0) {
                    if(specific == null) {
                        specific = entry;
                        distOld = distNew;
                    } else {
                        if(distNew < distOld) {
                            specific = entry;
                            distOld = distNew;
                        }
                    }
                }
                // Exact match.
                if(distNew == 0) break;
            }
            
            if(specific != null)
                return specific.exeFunction;

            if(anyBetter != null)
                return anyBetter.exeFunction;
            
            if(anyEntry != null)
                return anyEntry.exeFunction;
        }
        
        return null;
    }
    
    static class Entry {
        Class<?>[] paramTypes;
        Function exeFunction;
        
        Entry(Function f) {
            exeFunction = f;
            paramTypes = f.getParamTypes();
        }
        
        boolean anyParams() {
            if(paramTypes == null) return true;
            else return false;
        }
        
        boolean endsWithAnyParams() {
            if(paramTypes == null) return true;
            
            // The last parameter type is null.
            if(paramTypes.length > 0 &&
               paramTypes[paramTypes.length-1] == null) {
                return true;
            }
            
            return false;
        }
        
        /**
         * Example:
         *   argTypes:   {X.class, Y.class, Z.class}
         *   paramTypes: null
         *   result:     {X.class, Y.class, Z.class}
         *   
         * Example:
         *   argTypes:   {X.class, Y.class, Z.class}
         *   paramTypes: {null}
         *   result:     {X.class, Y.class, Z.class}
         *   
         * Example:
         *   argTypes:   {X.class, Y.class, Z.class}
         *   paramTypes: {T.class, null}
         *   result:     {T.class, Y.class, Z.class}
         *   
         * Example:
         *   argTypes:   {X.class, Y.class}
         *   paramTypes: {T.class, null}
         *   result:     {T.class, Y.class}
         *   
         * Example:
         *   argTypes:   {X.class}
         *   paramTypes: {T.class, null}
         *   result:     {T.class}
         *   
         * Example:
         *   argTypes:   null
         *   paramTypes: {T.class, null}
         *   result:     null
         */
        Class<?>[] getExtendedParamTypes(Class<?>[] argTypes) {
            Class<?>[] result = null;
            
            if (!endsWithAnyParams()) {
                return result;
            }
            
            int copyCount = 0;
            if(paramTypes != null) {
                copyCount = paramTypes.length - 1;
            }
            
            int numParamTypes = argTypes.length;
            if (numParamTypes >= copyCount) {
                result = new Class[numParamTypes];
                for(int i = 0; i < numParamTypes; i++) {
                    if(i < copyCount) {
                        result[i] = paramTypes[i];
                    } else {
                        result[i] = argTypes[i];
                    }
                }
            }
            
            return result;
        }
    }
    
    static class EntryContainer {
        ArrayList<Entry> list;
        
        EntryContainer() {
            list = new ArrayList<Entry>(3);
        }
        
        void add(Entry entry) {
            list.add(entry);
        }
    }
}
